#version 150 core
#extension GL_ARB_explicit_attrib_location : enable

layout( location = 0 ) in vec3 inVertex;
layout( location = 1 ) in vec3 inTextureCoord;
layout( location = 2 ) in vec3 inNormal;
layout( location = 3 ) in vec3 inTangent;
layout( location = 4 ) in vec3 inBoneId;
layout( location = 5 ) in vec3 inWeight;

layout( std140 ) uniform Parameters {
  mat4  viewProjectionMatrix;
  vec3  screenGamma;
  vec2  screenSize;
  float exposure;
  float alphaTestThreshold;
};

uniform usamplerBuffer objectIndexBuffer;
uniform usamplerBuffer boneIndexBuffer;
uniform samplerBuffer  commonBuffer;
uniform usamplerBuffer materialIndexBuffer;
uniform usamplerBuffer materialInfoBuffer;
uniform samplerBuffer  textureInfoBuffer;

uniform bool isInterface;
uniform int  objectOffset;

out VertexData {
  flat bool  isFont;
  flat float fontSmoothing;
  flat bool  hasOutline;
  flat float outlineWidth;
  flat bool  shouldPerformScissorTest;
  flat vec4  scissorInfo;
  vec2       textureCoord;
  flat vec2  textureCoordScale;
  flat vec2  textureCoordOffset;
  flat bool  isColor;
  flat vec4  color;
  flat vec4  secondColor;
  flat float opacity;
  flat float selfIllumination;
  mat3       tbnMatrix;
}
VertexOut;

const int maxWeightCount = 2;

// начало дополнительных данных кости, располагающееся после информации о матрицах
const int boneDataOffset = 4;

mat4 getMatrix( in int matrixIndex ) {
  return mat4(                                    //
    texelFetch( commonBuffer, matrixIndex ),      //
    texelFetch( commonBuffer, matrixIndex + 1 ),  //
    texelFetch( commonBuffer, matrixIndex + 2 ),  //
    texelFetch( commonBuffer, matrixIndex + 3 )   //
  );
}

int replaceMaterialId( in int materialId, in int replacementInfoOffset ) {
  ivec4 materialReplacementInfo = ivec4( texelFetch( commonBuffer, replacementInfoOffset ) );
  if ( materialReplacementInfo.x == materialId ) {
    return int( texelFetch( commonBuffer, replacementInfoOffset + 1 ).x );
  }
  if ( materialReplacementInfo.y == materialId ) {
    return int( texelFetch( commonBuffer, replacementInfoOffset + 1 ).y );
  }
  if ( materialReplacementInfo.z == materialId ) {
    return int( texelFetch( commonBuffer, replacementInfoOffset + 1 ).z );
  }
  if ( materialReplacementInfo.w == materialId ) {
    return int( texelFetch( commonBuffer, replacementInfoOffset + 1 ).w );
  }
  return materialId;
}

void main() {
  // индекс текущего объекта в общем массиве индексов objectIndexBuffer
  ivec2 objectIndexInfo = ivec2( texelFetch( objectIndexBuffer, gl_InstanceID + objectOffset ).xy );

  ivec4 boneIndexInfo;
  if ( isInterface ) {
    boneIndexInfo = ivec4( texelFetch( boneIndexBuffer, objectIndexInfo.x + int( inBoneId[ 0 ] ) ) );
    // рисуем интерфейс
    vec2 ndcPosition = ( vec2( getMatrix( boneIndexInfo.x ) * vec4( inVertex, 1.0f ) ) / screenSize ) * 2.0f - 1.0f;
    gl_Position      = vec4( ndcPosition, 0.0f, 1.0f );
  } else {
    // модельная матрица объекта
    mat4 modelMatrix = getMatrix( objectIndexInfo.y );

    // если вершина имеет информацию о влиянии нескольких костей, необходимо учесть это влияние
    int weightCount = maxWeightCount;
    if ( inWeight[ 0 ] == 1.0f ) {
      weightCount = 1;
    }

    vec4 resultPosition = vec4( 0.0f );
    // обрабатываем веса в обратном порядке, чтобы в конце осталась информация о текущей кости
    for ( int i = weightCount - 1; i >= 0; --i ) {
      boneIndexInfo = ivec4( texelFetch( boneIndexBuffer, objectIndexInfo.x + int( inBoneId[ i ] ) ) );

      mat4 boneTransform = getMatrix( boneIndexInfo.x );
      resultPosition += boneTransform * vec4( inVertex, 1.0f ) * inWeight[ i ];
      // resultNormal += boneTransform * vec4( inNormal, 0.0f ) * inWeight[ i ];
      // resultTangent += boneTransform * vec4( inTangent, 0.0f ) * inWeight[ i ];
    }

    gl_Position = viewProjectionMatrix * modelMatrix * resultPosition;
  }

  // bone properties
  VertexOut.isFont     = false;
  VertexOut.hasOutline = false;
  if ( boneIndexInfo.y > 0 ) {
    vec2 boneProperties = texelFetch( commonBuffer, boneIndexInfo.x + boneDataOffset ).xy;
    if ( boneProperties.x != 0.0f ) {
      VertexOut.isFont        = true;
      VertexOut.fontSmoothing = boneProperties.x;
    }
    if ( boneProperties.y != 0.0f ) {
      VertexOut.hasOutline   = true;
      VertexOut.outlineWidth = boneProperties.y;
    }
  }

  // bone material replacement info
  int materialId = int( inTextureCoord.z );
  if ( boneIndexInfo.z > 0 ) {
    // передаем сдвиг для данных замены материала относительно исходной позиции + все предыдущие данные
    materialId = replaceMaterialId( materialId, boneIndexInfo.x + boneDataOffset + boneIndexInfo.y );
  }

  // scissor info
  VertexOut.shouldPerformScissorTest = false;
  if ( boneIndexInfo.w > 0 ) {
    VertexOut.shouldPerformScissorTest = true;
    int scissorOffset                  = boneIndexInfo.x + boneDataOffset + boneIndexInfo.y;
    if ( boneIndexInfo.z > 0 ) {
      // поскольку информация о заменяемых материалах состоит из двух векторов
      scissorOffset += 2;
    }
    VertexOut.scissorInfo = texelFetch( commonBuffer, scissorOffset );
  }

  // информация о материале: x - id, y - count
  VertexOut.textureCoord = inTextureCoord.xy;
  ivec2 materialIndex    = ivec2( texelFetch( materialIndexBuffer, materialId ).xy );
  if ( materialIndex.y > 0 ) {
    // первый пакет содержит информацию о картах albedo, normal, height
    ivec4 materialInfoFirst = ivec4( texelFetch( materialInfoBuffer, materialIndex.x ) );
    // albedo
    if ( materialInfoFirst.x > 0 ) {
      vec4 albedoInfo              = texelFetch( textureInfoBuffer, materialInfoFirst.x );
      VertexOut.textureCoordScale  = albedoInfo.xy;
      VertexOut.textureCoordOffset = albedoInfo.zw;
    }
  }

  VertexOut.isColor = false;
  if ( materialIndex.y > 1 ) {
    // второй пакет содержит информацию о цвете и дополнительные свойства: прозрачность, светимость
    ivec4 materialInfoSecond = ivec4( texelFetch( materialInfoBuffer, materialIndex.x + 1 ) );
    // color
    if ( materialInfoSecond.x > 0 ) {
      VertexOut.isColor = true;
      VertexOut.color   = texelFetch( textureInfoBuffer, materialInfoSecond.x );
    }
    // second color
    if ( materialInfoSecond.y > 0 ) {
      VertexOut.isColor     = true;
      VertexOut.secondColor = texelFetch( textureInfoBuffer, materialInfoSecond.y );
    }
    // advanced material info
    if ( materialInfoSecond.z > 0 ) {
      vec2 advancedInfo          = texelFetch( textureInfoBuffer, materialInfoSecond.z ).xy;
      VertexOut.opacity          = advancedInfo.x;
      VertexOut.selfIllumination = advancedInfo.y;
    }
  }

  // // создаем матрицу перевода направления в координаты пространства мира
  // mat3 normalMatrix = normalMatrixList[matrixIndex];
  // vec3 wsNormal    = normalize( normalMatrix * inNormal );
  // vec3 wsTangent   = normalize( normalMatrix * inTangent );
  // vec3 wsBitangent = normalize( cross( wsNormal, wsTangent ) );
  // tbnMatrix        = mat3( wsTangent, wsBitangent, wsNormal );
  VertexOut.tbnMatrix = mat3( 1.0f );
}
