#version 150 core
#extension GL_ARB_explicit_attrib_location : enable

layout( std140 ) uniform Matrices {
  mat4 viewProjectionMatrix;
  mat4 viewProjectionInversedMatrix;
};

in vec2  textureCoord;
in mat3  tbnMatrix;
in mat4  tangentToClipMatrix;
in vec3  wsPosition;
in vec3  tsPosition;
in vec3  tsCameraPosition;
in float distanceToFragment;

uniform sampler2D textureAlbedo;
uniform sampler2D textureNormal;
uniform sampler2D textureHeight;

uniform bool  drawPOM;
uniform float alphaTestThreshold;
uniform bool  homogeneousReversedDepth;
uniform vec3  wsCameraPosition;
uniform vec2  screenSize;

// переданные параметры полигона
uniform bool  isColor;
uniform vec4  color;
uniform float opacity;
uniform float selfIllumination;

// Ouput data
layout( location = 0 ) out vec4 albedoBuffer;
layout( location = 1 ) out vec4 normalBuffer;

// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec4 getWSPosition( in vec2 ssPosition, in float z ) {
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    z = z * 2.0f - 1.0f;
  }
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssPosition * 2.0f - 1.0f, z, 1.0f );
  return wsPosition / wsPosition.w;
}

float getHeight( in vec2 texCoord ) {
  return 1.0f - texture( textureHeight, texCoord ).x;
}
// возвращает величину сдвига в tangent space относительно исходных текстурных координат
vec3 parallaxOcclusionMapping( in vec2 texCoord ) {
  float currentHeightMapValue = getHeight( texCoord );
  // POM имеет смысл считать только для фрагментов с ненулевой высотой
  if ( currentHeightMapValue <= 0.03f ) {
    return vec3( 0.0f, 0.0f, 0.0f );
  }

  vec3  cameraDirection = normalize( tsPosition - tsCameraPosition );
  float heightScale     = 0.01f;
  // вычисляем количество слоев, в зависимости от угла наклона камеры относительно поверхности
  const float minLayers = 16;
  const float maxLayers = 32;
  int         numLayers = int( mix( maxLayers, minLayers, abs( dot( vec3( 0.0f, 0.0f, 1.0f ), cameraDirection ) ) ) );
  // высота одного слоя
  float layerHeight = 1.0f / numLayers;
  // величина сдвига текстурных координат для каждого слоя
  vec2 deltaTexCoords = layerHeight * heightScale * ( cameraDirection.xy / cameraDirection.z );

  vec2  currentTexCoords   = texCoord;
  float currentLayerHeight = 0.0f;
  vec2  prevTexCoords      = currentTexCoords;
  float prevHeightMapValue = currentHeightMapValue;
  while ( currentLayerHeight < currentHeightMapValue ) {
    // обновляем значения предыдущего шага
    prevTexCoords      = currentTexCoords;
    prevHeightMapValue = currentHeightMapValue;
    // высота текущего слоя
    currentLayerHeight += layerHeight;
    // сдвигаем текстурные координаты
    currentTexCoords -= deltaTexCoords;
    // новое значение высоты
    currentHeightMapValue = getHeight( currentTexCoords );
  }
  // считаем высоты текущего и предыдущего слоя и делаем линейную интерполяцию текстурных координат
  float currentHeight  = currentHeightMapValue - currentLayerHeight;
  float previousHeight = prevHeightMapValue - currentLayerHeight + layerHeight;
  float weight         = currentHeight / ( currentHeight - previousHeight );
  currentTexCoords     = prevTexCoords * weight + currentTexCoords * ( 1.0f - weight );
  // интерполяция высоты
  float interpolatedHeight = currentLayerHeight + previousHeight * weight + currentHeight * ( 1.0f - weight );

  // TODO: подумать, можно ли тут избавиться от вычитания, сразу посчитав величину сдвига
  return vec3( texCoord - currentTexCoords, interpolatedHeight * 0.1f );
}

vec3 getNormal( in vec2 texCoord ) {
  vec3 tsNormal = texture( textureNormal, texCoord ).xyz;
  // переводим в NDC [-1.0f, 1.0f]
  vec3 ndcNormal = tsNormal * 2.0f - 1.0f;
  // переводим в мировое пространство
  return normalize( tbnMatrix * ndcNormal );
}

void main() {
  const float drawPOMDistance = 45.0f;
  vec3        tsShift         = vec3( 0.0f, 0.0f, 0.0f );
  vec2        texCoord        = textureCoord;
  if ( drawPOM && distanceToFragment < drawPOMDistance ) {
    // получаем смещение текстурных координат
    tsShift = parallaxOcclusionMapping( texCoord );
    texCoord -= tsShift.xy;
  }

  // записываем светимость объекта 4-м компонентом в карту нормалей
  normalBuffer = vec4( getNormal( texCoord ), selfIllumination );

  if ( isColor ) {
    albedoBuffer = color;
    return;
  }

  vec4 albedoSample = texture( textureAlbedo, texCoord );

  float alpha = min( albedoSample.a, opacity );
  if ( alpha < alphaTestThreshold ) {
    discard;
  }

  albedoBuffer = vec4( albedoSample.rgb, alpha );
}
