#version 150 core
#extension GL_ARB_explicit_attrib_location : enable

layout( std140 ) uniform Matrices {
  mat4 viewProjectionMatrix;
  mat4 viewProjectionInversedMatrix;
};

uniform sampler2D normalBuffer;
uniform sampler2D depthBuffer;

uniform vec2 screenSize;
uniform vec3 wsCameraPosition;

uniform bool homogeneousReversedDepth;

uniform int numSamples;
uniform int numSpiralTurns;

uniform float projectedRadius;

/** World-space AO radius in scene units (r).  e.g., 1.0m */
uniform float radius;
uniform float negInvRadius2;

/** Bias to avoid AO in smooth corners, e.g., 0.01m */
uniform float bias;
uniform float intensity;

// Ouput data
layout( location = 0 ) out float visibility;

// получаем однородные координаты экрана из координат окна
vec2 getSSPosition( in vec2 windowSpacePosition ) {
  return windowSpacePosition / screenSize;
}
// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec4 getWSPosition( in vec2 ssPosition, in float z ) {
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    z = z * 2.0f - 1.0f;
  }
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssPosition * 2.0f - 1.0f, z, 1.0f );
  return wsPosition / wsPosition.w;
}

/** Returns a unit vector and a screen-space radius for the tap on a unit disk
    (the caller should scale by the actual disk radius) */
vec2 tapLocation( int sampleNumber, float spinAngle, out float ssRadius ) {
  // Radius relative to ssRadius
  float radius = float( sampleNumber + mod( spinAngle, 1.0f ) + 0.5f ) * ( 1.0f / ( numSamples - 0.5f ) );
  radius *= radius;
  float angle = radius * ( numSpiralTurns * 6.28f ) + spinAngle;

  ssRadius = radius;
  return vec2( cos( angle ), sin( angle ) );
}

/** Smaller return value = less occlusion */
float fallOffFunction( float vv, float vn, float epsilon ) {
  float f = clamp( vv * negInvRadius2 + 1.0f, 0.0f, 1.0f );
  return f * clamp( vn - bias, 0.0f, 1.0f ) / ( epsilon + vv );
}

/** Compute the occlusion due to sample point \a wsSceneSample about camera-space point \a wsPosition with unit normal \a wsNormal */
float aoValueFromPositionsAndNormal( vec3 wsPosition, vec3 wsNormal, vec3 wsSceneSample ) {
  vec3        v       = wsSceneSample - wsPosition;
  float       vv      = dot( v, v );
  float       vn      = dot( v, wsNormal );
  const float epsilon = 0.1f;

  float f = fallOffFunction( vv, vn, epsilon );

  // Enhance dark areas
  return f * mix( 0.9f + 0.5f * intensity, 1.2f - intensity * 0.15f, f );
}

/** Compute the occlusion due to sample with index \a i about the pixel at \a ssCoord that corresponds
	to camera-space point \a wsPosition with unit normal \a wsNormal

	Note that units of H() in the HPG12 paper are meters, not
	unitless.  The whole falloff/sampling function is therefore
	unitless.  In this implementation, we factor out (9 / radius).
*/
float sampleAO( in vec3 wsPosition, in vec3 wsNormal, in int sampleIndex, in float randomPatternRotationAngle, in mat3 tbnMatrix ) {
  /////////////////////////
  // улучшеный вариант алгоритма SAO
  float ssRadius;
  vec2  unitOffset = tapLocation( sampleIndex, randomPatternRotationAngle, ssRadius );

  // выбираем смещенный семпл методом hemisphere
  vec3 tsSamplePosition = ssRadius * vec3( unitOffset, ssRadius );
  vec3 wsSamplePosition = tbnMatrix * tsSamplePosition;

  wsSamplePosition = wsPosition + radius * wsSamplePosition;

  // project sample position (to sample texture) (to get position on screen/texture)
  vec4 offset = viewProjectionMatrix * vec4( wsSamplePosition, 1.0f );  // from world to clip-space
  offset.xy /= offset.w;                                                // perspective divide
  offset.xy = offset.xy * 0.5 + 0.5;                                    // transform to range 0.0 - 1.0

  // если мип не считать и всегда брать из нулевого, получается быстрее всего, качество не страдает
  vec3 wsSceneSample = getWSPosition( offset.xy, textureLod( depthBuffer, offset.xy, 0 ).x ).xyz;

  return aoValueFromPositionsAndNormal( wsPosition, wsNormal, wsSceneSample );
}

const float MIN_RADIUS = 10.0f;  // pixels

float lowPrecisionHash( vec2 p ) {
  vec3 p3 = fract( vec3( p.xyx ) * 0.13f );
  p3 += dot( p3, p3.yzx + 3.333f );
  return fract( ( p3.x + p3.y ) * p3.z ) * 20.1f;
}

float pixelPositionHash( ivec2 p ) {
  return float(
           ( ( p.x & 0xffff ) * ( ( p.y & 0xff ) + 1 ) ) ^ ( p.x >> 16 ) + ( ( ( p.x + 127 ) & 0xffff ) * ( ( p.y + 128 ) & 0xff ) ) ) *
    2.39996f;
}

void main() {
  ivec2 ssCoord  = ivec2( gl_FragCoord.xy );
  vec2  texCoord = gl_FragCoord.xy / screenSize;

  // World space point being shaded
  float csDepth    = textureLod( depthBuffer, texCoord, 0 ).x;
  vec3  wsPosition = getWSPosition( texCoord, csDepth ).xyz;

  // расстояние от камеры до фрагмента сцены
  float distanceToFragment = length( wsPosition - wsCameraPosition );
  // Choose the screen-space sample radius
  // proportional to the projected area of the sphere
  float ssDiskRadius = projectedRadius / distanceToFragment;
  if ( ssDiskRadius <= MIN_RADIUS ) {
    // There is no way to compute AO at this radius
    visibility = 1.0;
    return;
  }

  vec3 wsNormal = textureLod( normalBuffer, texCoord, 0 ).xyz;

  // create tbnMatrix change-of-basis matrix: from tangent-space to world-space
  vec3 randomVec = normalize( wsCameraPosition - wsPosition );
  vec3 tangent   = normalize( randomVec - wsNormal * dot( randomVec, wsNormal ) );
  vec3 bitangent = cross( wsNormal, tangent );
  mat3 tbnMatrix = mat3( tangent, bitangent, wsNormal );

  // Hash of pixel position as pseudo-random value
  //float randomPatternRotationAngle = pixelPositionHash( ssCoord );
  float randomPatternRotationAngle = lowPrecisionHash( gl_FragCoord.xy );

  float sum = 0.0f;
  for ( int i = 0; i < numSamples; ++i ) {
    sum += sampleAO( wsPosition, wsNormal, i, randomPatternRotationAngle, tbnMatrix );
  }

  visibility = pow( clamp( 1.0f - sqrt( sum * ( 1.0f / numSamples ) ), 0.0f, 1.0f ), intensity );
}
